import logging
import time
import os
import math
import pandas as pd

try:
    from spswarehouse.credentials import powerschool_config
except ModuleNotFoundError:
    powerschool_config = None
    print("No credentials file found in spswarehouse. This could cause issues.")

from ducttape.utils import (
    DriverBuilder,
    get_most_recent_file_in_dir,
)

from spswarehouse.general.selenium import (
    type_in_element_by_id,
    select_visible_text_in_element_by_id,
    click_element_by_id,
    click_element_by_name,
    click_element_by_partial_link_text,
    wait_for_element_containing_specific_text,
)
    
from selenium.webdriver.support.ui import WebDriverWait, Select
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.action_chains import ActionChains
from selenium.common.exceptions import NoSuchElementException

ADMIN_LOGIN_PAGE_PATH = 'admin/pw.html'
ADMIN_HOME_PAGE_PATH = 'admin/home.html'
ADMIN_URL_SCHEME = 'https://'
STATE_REPORTS_PAGE_PATH = 'admin/reports/statereports.html?repType=state'
REPORT_QUEUE_REPORTWORKS_PAGE_PATH = 'admin/reportqueue/prhome.html'
REPORT_QUEUE_SYSTEM_PAGE_PATH = 'admin/reportqueue/home.html'
DATA_IMPORT_MANAGER_PATH = 'admin/datamgmt/importmanager.action'

DEPRECATION_WARNING_MESSAGE = "DEPRECATION_WARNING: This helper function in the PowerSchool class will be deprecated in a future release. Use the helper functions in general.selenium instead."

class PowerSchool:
    """
    This class is an abstraction for interacting with the PowerSchool Admin user 
    interface via Selenium.
    """

    def __init__(
        self,
        config: dict=None,
        username: str=None,
        password: str=None,
        host: str=None,
        headless: bool=True, 
        download_location: str='.',
        chrome_option_prefs: dict=None,
    ):
        
        if config is None:
            config = powerschool_config
        
        if username is None: 
            username = config['username']
        else:
            username = username

        if password is None:
            password = config['password']
        else:
            password = password

        if host is None:
            self.host = config['host']
        else:
            self.host = host

        self.driver = DriverBuilder().get_driver(
            headless=headless,
            download_location=download_location,
            chrome_option_prefs=chrome_option_prefs,
        )
        
        self._log_into_powerschool_admin(username, password)

    def quit(self):
        self.driver.quit()

    def refresh(self):
        self.driver.refresh()

    def ensure_on_desired_path(self, desired_path: str):
        """
        Checks whether the WebDriver is on the desired path. If not, navigates there.
        When using this function, consider a WebDriverWait afterwards to confirm the
        desired page has loaded.

        Parameters:
        self
        desired_path: The path to be checked against

        Returns:
        n/a
        """

        current_path = self._get_current_path()

        logging.info(f"The current path is: {current_path}")

        if(current_path == desired_path):
            logging.info("The current path is the desired path. No action taken.")
        else:
            logging.info(f"This does not match {desired_path}, so going to that path")
            self.driver.get('https://' + self._get_current_domain() + "/" + desired_path)
            time.sleep(3) # Give new page time to load
            logging.info(f"Moved to {desired_path}.")

    def check_whether_desired_school_selected(self, school_name: str) -> bool:
        """
        Checks whether the specified school is currently selected in PowerSchool but
        takes no action either way.

        Parameters:
        self
        school_name: The school name for selecting from the drop-down in the upper-right of the user 
        interface. Should be the complete school name to avoid incorrect partial matches.

        Returns:
        bool: True indicates the desired school is selected, and False indicates it is not.
        """

        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.ID, 
            'school_picker_adminSchoolPicker_toggle_btn')))

        elem.click()
        time.sleep(1)

        selected_element = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((
            By.CSS_SELECTOR, '.list-item.selectable.selected')))

        school_element_text = selected_element.find_element(By.XPATH, ".//div").text

        result_message = f"Found the {school_name} in the school element text. Match!"
        outcome = True

        if(school_name not in school_element_text):
            result_message = f"Did not find {school_name} in the school element text. No match."
            outcome = False

        logging.info(result_message)

        logging.info(f"Pressing escape to leave dropdown selection.")
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ESCAPE).perform()
        
        return outcome
    
    def check_whether_desired_school_year_selected(self, school_year_dropdown: str):
        """
        Checks whether the specified school year is currently selected in PowerSchool but
        takes no action either way.

        Parameters:
        self
        school_year_dropdown: The exact text of the school year that appears in the dropdown in 
            PowerSchool. Should be 'XX-YY 20XX-20YY' format.

        Returns:
        bool: True indicates the desired school year is selected, and False indicates it is not.
        """

        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.ID, 
            'term_picker_adminTermPicker_toggle_btn')))

        elem.click()
        time.sleep(1)

        selected_element = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((
            By.CSS_SELECTOR, '.list-item.selectable.selected')))

        school_year_element_text = selected_element.find_element(By.XPATH, ".//div").text

        result_message = f"Found '{school_year_dropdown}' in the school year element text. Match!"
        outcome = True

        if(school_year_dropdown not in school_year_element_text):
            result_message = f"Did not find '{school_year_dropdown}' in the school year element text. No match."
            outcome = False

        logging.info(result_message)

        logging.info(f"Pressing escape to leave dropdown selection.")
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ESCAPE).perform()
        
        return outcome
    
    def switch_to_school_year(self, school_year_dropdown: str):
        """
        Switches to a specified school year in PowerSchool.

        Parameters:
        self
        school_year_dropdown: The exact text of the school year that appears in the dropdown in 
            PowerSchool. Should be 'XX-YY 20XX-20YY' format.

        Returns:
        n/a
        """
        if self.check_whether_desired_school_year_selected(school_year_dropdown) == False:
            logging.info(f"'{school_year_dropdown}' is not already selected. Selecting now.")

            self.ensure_on_desired_path(ADMIN_HOME_PAGE_PATH)

            logging.info("Waiting for School Year Picker")
            elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.ID, 
                'term_picker_adminTermPicker_toggle_btn')))
            logging.info("School Year Picker found. Click it.")

            elem.click()

            time.sleep(1)

            logging.info("Waiting for School Year Search Field")
            elem = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.ID, 
                'termText')))
            logging.info("Found School Year Search Field. Typing in school year text.")

            # Only send the XX-YY portion of the dropdown text
            elem.send_keys(school_year_dropdown[:5])

            time.sleep(1)

            logging.info("Looking for first school year in list")
            
            elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.XPATH, 
                "//ul[@id='term_choices']/li[2]"))) # This is li[2] instead of li[1] because there is read-only school year item displayed

            logging.info("Found first school year in results list. Clicking.")
            elem.click()
            logging.info("Click. Waiting for page to refresh.")
            time.sleep(1)
            
            assert self.check_whether_desired_school_year_selected(school_year_dropdown), "Failed to select \
                desired school year."
        else:
            logging.info(f"'{school_year_dropdown}' is already selected. No action taken.")

    def switch_to_school(self, school_name: str):
        """
        Switches to a specified school in PowerSchool.

        Parameters:
        self
        school_name: The school name for selecting from the drop-down in the upper-right of the user 
            interface. Should be the complete school name to avoid incorrect partial matches.

        Returns:
        n/a
        """

        if self.check_whether_desired_school_selected(school_name) == False:
            logging.info(f"{school_name} is not already selected. Selecting now.")

            self.ensure_on_desired_path(ADMIN_HOME_PAGE_PATH)

            logging.info("Waiting for School Picker")
            elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.ID, 
                'school_picker_adminSchoolPicker_toggle_btn')))
            logging.info("School Picker found. Click it.")

            elem.click()

            time.sleep(1)

            logging.info("Waiting for School Search Field")
            elem = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.ID, 
                'schoolSearchField_value')))
            logging.info("Found School Search Field. Typing in school name.")

            elem.send_keys(school_name)

            time.sleep(1)

            logging.info("Looking for first school in list")
            elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.XPATH, 
                "//ul[@id='school_choices']/li[1]")))

            logging.info("Found first school in results list. Clicking.")
            elem.click()
            logging.info("Click. Waiting for page to refresh.")
            time.sleep(1)
            
            assert self.check_whether_desired_school_selected(school_name), "Failed to select \
                desired school."
        else:
            logging.info(f"{school_name} is already selected. No action taken.")
    
    def navigate_to_state_reports_page(self):
        """
        Navigates to the state reports page in PowerSchool.

        Parameters:
        self
        
        Returns:
        n/a
        """

        self.driver.get(ADMIN_URL_SCHEME + self._get_current_domain() + '/' + STATE_REPORTS_PAGE_PATH)

        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.CSS_SELECTOR, 
            'ul li.selected')))
        assert elem.text == 'State', "'State' tab is not selected"

    def navigate_to_specific_state_report(self, report_link_text: str):
        """
        Navigates to the state reports page in PowerSchool and clicks on the report containing
        the report_link_text string.

        Parameters:
        self
        report_link_text: The string to uniquely identify the report. Can be partial text.

        Returns:
        n/a
        """
        self.navigate_to_state_reports_page()

        click_element_by_partial_link_text(self.driver, report_link_text)



### START TODO: These helper functions inside the PowerSchool class will be removed in a future release. ##############
###             The functions in general.selenium should be used instead. #############################################

    def helper_type_in_element_by_id(self, element_id: str, input_to_type: str):
        """
        Waits for an element by ID, clears it, and types in the input.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.ID, element_id)))
        elem.clear()
        elem.send_keys(input_to_type)

    def helper_type_in_element_by_name(self, element_name: str, input_to_type: str):
        """
        Waits for an element by name, clears it, and types in the input.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.NAME, 
            element_name)))
        elem.clear()
        elem.send_keys(input_to_type)

    def helper_select_visible_text_in_element_by_id(self, element_id: str, 
        text_to_select: str):
        """
        Waits for an element by ID and selects it by specified text.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.ID, element_id)))
        select = Select(elem)
        select.select_by_visible_text(text_to_select)

    def helper_select_visible_text_in_element_by_name(self, element_name: str, 
        text_to_select: str):
        """
        Waits for an element by name and selects it by specified text.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.NAME, 
            element_name)))
        select = Select(elem)
        select.select_by_visible_text(text_to_select)

    def helper_click_element_by_id(self, element_id: str):
        """
        Waits for an element by ID and clicks it.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.ID, element_id)))
        elem.click()

    def helper_click_element_by_name(self, element_name: str):
        """
        Waits for an element by name and clicks it.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.NAME, element_name)))
        elem.click()

    def helper_click_element_by_partial_link_text(self, partial_link_text: str):
        """
        Waits for an element by partial link text and clicks it.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.PARTIAL_LINK_TEXT, 
            f"{partial_link_text}")))
        elem.click()

    def helper_ensure_checkbox_is_checked_by_name(self, checkbox_name: str):
        """
        Waits for a checkbox element by name and clicks it if it is not already selected.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        checkbox = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.XPATH, 
            f"//input[@type='checkbox' and @name='{checkbox_name}']")))
        
        if checkbox.is_selected() == False:
            checkbox.click()
    
    def helper_ensure_checkbox_is_unchecked_by_name(self, checkbox_name: str):
        """
        Waits for a checkbox element by name and clicks it if it is already selected, to make
        sure it is not checked.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        checkbox = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.XPATH, 
            f"//input[@type='checkbox' and @name='{checkbox_name}']")))
        
        if checkbox.is_selected():
            checkbox.click()

    def helper_ensure_element_text_matches_expected_value_by_xpath(self, element_xpath, expected_text):
        """
        Waits for an element by XPATH and checks whether its text matches the expected text.
        """
        print(DEPRECATION_WARNING_MESSAGE)

        elem = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.XPATH, 
            element_xpath)))
        
        return elem.text == expected_text
    
    def helper_wait_for_element_containing_specific_text(self, expected_element_text, wait_time_in_seconds=30):
        """
        Waits for an element containing specific text raises an exception if it does not 
        appear in the time allotted (default = 30 seconds).
        """
        print(DEPRECATION_WARNING_MESSAGE)

        try:
            WebDriverWait(self.driver, wait_time_in_seconds).until(EC.presence_of_element_located((By.XPATH, f"//*[contains(text(), '{expected_element_text}')]")))
        except:
            raise Exception(f'Element with text "{expected_element_text}" not found within {wait_time_in_seconds} seconds.')

### END TODO #########################################################################################################



    def download_latest_report_from_report_queue_reportworks(self, destination_directory_path: str = '.', 
        file_postfix: str = ''):
        """
        Navigates to the PowerSchool Report Queue (ReportWorks), confirms the most recent report is 
        done generating, and downloads it.

        Parameters:
        self
        destination_directory_path: Where to download the PowerSchool report to.
        file_postfix: Optional postfix to attach to the end of the downloaded file's filename.

        Returns:
        bool: True once successfully downloads the report. Otherwise, function keeps looping.
        """
        self.ensure_on_desired_path(REPORT_QUEUE_REPORTWORKS_PAGE_PATH)

        # Pause briefly to give the just-submitted report time to get into the queue
        time.sleep(5)

        # Confirm that the page has loaded
        elem = WebDriverWait(self.driver, 5).until(EC.element_to_be_clickable((By.ID, 'prReloadButton')))
        elem.click()
        logging.info('Refresh button on report page has loaded. Refreshing.')
        
        while True:
            # TODO: Add a counter so this function can't get stuck in an infinte loop.
            try:
                # Confirm no reports are running
                self.driver.find_element(By.XPATH, "//p[contains(text(), 'No reports running or pending!')]")
                
                # TODO: Is the below refresh necessary when we're refreshing upon first hitting this page?
                # There is occasional flakiness where the "No reports running or pending!" message 
                #    shows up but the latest report is not in the list for downloading yet, so refresh 
                #    the page one more time.
                time.sleep(1)
                elem = WebDriverWait(self.driver, 5).until(EC.element_to_be_clickable((By.ID, 'prReloadButton')))
                elem.click()
                time.sleep(1)
            except NoSuchElementException: # Because reports ARE running
                time.sleep(5)
                elem = WebDriverWait(self.driver, 5).until(EC.element_to_be_clickable((By.ID, 'prReloadButton')))
                elem.click()
                logging.info('PowerSchool report is not ready, refreshing and waiting.')
                time.sleep(3)
            else:
                # Download the first report in table
                queued_reports = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((
                    By.XPATH, '//*[@id="queuecontent"]/table/tbody/tr[2]/td[8]/a')))
                download_link = queued_reports.get_attribute('href')
                original_files_list = os.listdir(destination_directory_path)
                self.driver.get(download_link) #downloads the file
                logging.info('PowerSchool report downloaded.')
                break

        self._wait_for_new_file_in_folder(destination_directory_path, original_files_list)
        self._rename_recent_file_in_dir(destination_directory_path, file_postfix)
        logging.info('Successfully renamed the downloaded file.')

        return True

    def download_latest_report_from_report_queue_system(self, destination_directory_path: str = '.', 
        file_postfix: str = ''):
        """
        Navigates to the PowerSchool Report Queue (System), confirms the most recent report is done 
        generating, and downloads it.

        Parameters:
        driver: A Selenium WebDriver
        destination_directory_path: Where to download the PowerSchool report to.
        file_postfix: Optional postfix to attach to the end of the downloaded file's filename.

        Returns:
        bool: True if successfully downloads a report, or False if it cannot, either because the 
            report generated no results from the previously-submitted parameters or the report
            download page is in a format this function does not handle.
        """
        self.ensure_on_desired_path(REPORT_QUEUE_SYSTEM_PAGE_PATH)

        # Pause briefly to give the just-submitted report time to get into the queue
        time.sleep(5)

        # Confirm that the page has loaded
        elem = WebDriverWait(self.driver, 5).until(EC.element_to_be_clickable((By.ID, 'prReloadButton')))
        elem.click()
        logging.info('Refresh button on report page has loaded. Refreshing.')

        while True:
            try:
                # Try to find a running report
                self.driver.find_element(By.XPATH, "//td[text()='Running']")

                # If yes, keep going here
                time.sleep(5)
                elem = WebDriverWait(self.driver, 5).until(EC.element_to_be_clickable((By.ID, 'prReloadButton')))
                elem.click()
                logging.info('PowerSchool report is not ready, refreshing and waiting.')
                time.sleep(3)
            except NoSuchElementException: # Because all reports are done running
                logging.info('No currently running reports. Downloading the most recently completed report.')
                break

        top_completed_report_view_link = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.XPATH,
            "//*[@id='content-main']/div[3]/table/tbody/tr[1]/td[a[text()='View']][1]/a"))) 
            # Note: The above XPATH seems to change a lot as PowerSchool makes changes to which column the link is in.
            # This is the hard-coded XPATH for the 6th column: '//*[@id="content-main"]/div[3]/table/tbody/tr[1]/td[6]/a'
            # The above XPATH tries to dynamically determine which column the link is in to be robust to these changes.
        top_completed_report_view_link.click()

        try:
            # Look for a result file link
            logging.info('Looking for the result file link.')
            download_link = WebDriverWait(self.driver, 10).until(EC.element_to_be_clickable((
                By.LINK_TEXT, 'Click to Download Result File')))
            original_files_list = os.listdir(destination_directory_path)
            download_link.click()
            logging.info('Downloading PowerSchool report.')

            self._wait_for_new_file_in_folder(destination_directory_path, original_files_list)
            self._rename_recent_file_in_dir(destination_directory_path, file_postfix)
            logging.info('Successfully renamed the downloaded file.')

            return True
        except:
            try:
                # If no result file link is found, look for a confirmation that no file was generated
                WebDriverWait(self.driver, 10).until(EC.element_to_be_clickable((By.XPATH, 
                    "//h1[text()='No records found']")))
                logging.info('PowerSchool reports "No records found"')
                return False
            except:
                # TODO: Check for the SCSC validation errors page and download it appropriately
                logging.info(f"Unable to confirm results. Please check manually for postfix \
                    {file_postfix}.")
                self.driver.back()
                return False
    
    def upload_student_csv_quick_import(self, filename):
        """
        Uploads a file to Quick Import. Requires your file to
        1. be a CSV
        2. have a header row
        3. have header names that can be auto-matched by PowerSchool
        4. have a column called "student_number"
        """
        
        import_data = pd.read_csv(filename)
        final_student_number = import_data.iloc[-1]['student_number']
        
        self.upload_csv_quick_import(filename, final_student_number, 'Students')
        
    def upload_csv_quick_import(self, filename, final_value, table_name):
        """
        Uploads a CSV file to quick import.
        
        Arguments:
        filename: Path to the file.
        final_value: A value in the final row of the file that can be used to identify
            when PS is done processing the file.
        table_name: Exact value of the table name from the PS UI
        """
        
        # Convert the file to a tab-delimited file
        import_data = pd.read_csv(filename, encoding='mac_roman')
        
        new_filename = filename + ".tsv"
        
        import_data.to_csv(
            new_filename,
            index=False,
            sep='\t',
            encoding='mac_roman',
            # The "CR" option in PS's quick import, which is the default, is in fact "\n"
            lineterminator = "\r",
        )
        
        self.upload_quick_import(new_filename, final_value, table_name)
        
    def upload_quick_import(self, filename, final_value, table_name):
        """
        Uploads a tab-delimited file to quick import.
        
        Arguments:
        filename: Path to the file.
        final_value: A value in the final row of the file that can be used to identify
            when PS is done processing the file.
        """
        
        # Upload may cover multiple schools, so should be done at the District Office level
        self.switch_to_school('District Office')

        # Go to Quick Import page
        self.ensure_on_desired_path('admin/importexport/quickimport/quickimport1.html')

        # Upload to designated table
        logging.info(f"Selecting {table_name} for table")
        select_visible_text_in_element_by_id(self.driver, 'filenumber', table_name)
        
        # Choose file to upload
        type_in_element_by_id(self.driver, 'filename', filename)

        # Submit file
        click_element_by_id(self.driver, 'btnImport')

        # Choose "Check to exclude first row"
        click_element_by_name(self.driver, 'skipFirstRow')

        # Choose "Update the student's record with the information from the file being imported."
        click_element_by_id(self.driver, 'rdioc_update')

        # Submit
        click_element_by_id(self.driver, 'btnSubmit')
        logging.info("Submitting file")

        # Check that file finished processing
        logging.info(f'Waiting for student ID #{final_value} to appear to indicate that the file is finished processing.')
        wait_for_element_containing_specific_text(self.driver, final_value, 60)
        logging.info('Final student found. Upload file finished processing.')
        
    def upload_data_import_manager(self, file_path, table_name, max_processing_wait_time_in_seconds = 60, 
        override_existing_record = False):
        """
        Uploads a tab-delimited file to the Data Import Manager.
        
        Arguments:
        file_path: Path to the tab-delimited file.
        table_name: The name of the PowerSchool table to upload to.
        max_processing_wait_time_in_seconds: The maximum number of seconds to wait for the file to processing.
            The processing check happens every 10 seconds, so this number will be divided by 10 and rounded down
            to determine how many checks to make. e.g., a value of 35 will check 3 times, which is roughly 30 seconds.
        override_existing_record: Defaults to False. Submit as True to have the function select the option to override
            existing records as part of the import. WARNING: Be very careful when using this feature, since overriding 
            records in PowerSchool requires very precisely formatted files, and if anything is incorrect, the system
            will likely create duplicate records. Please extensively test the file you are importing.
        """

        df_row_count = pd.read_csv(file_path)
        num_data_rows = df_row_count.shape[0]

        if num_data_rows == 0:
            raise Exception(f'Upload file at path "{file_path}" does not contain any rows of data.')

        num_rows_in_file = num_data_rows + 1 # Add 1 to account for the header row

        logging.info('Switching to the "District Office" for a Data Import Manager upload.')
        self.switch_to_school('District Office')

        self.ensure_on_desired_path(DATA_IMPORT_MANAGER_PATH)

        logging.info('Choosing the upload file.')
        type_in_element_by_id(self.driver, 'idFilename', file_path)

        logging.info('Selecting the upload table.')
        select_visible_text_in_element_by_id(self.driver, 'moduleSelect', table_name)
        
        # Need to give the UI a moment to catch up to the fact that a table has been selected.
        # (The button is already clickable when a table is unselected, but gives an error)
        time.sleep(5)
        logging.info('Clicking Next')
        click_element_by_id(self.driver, 'nextButton0')

        # Brief pause to allow for loading next part of the screen
        time.sleep(10)

        logging.info('Assuming all fields mapped properly.')
        click_element_by_id(self.driver, 'nextButton1')

        # WARNING: Be VERY careful when overriding existing records, because PowerSchool can create duplicates if the upload
        #   file is not precisely what the system expects.
        if override_existing_record:
            logging.info('Because override_existing_record was specified as True, clicking the corresponding radio button.')
            click_element_by_id(self.driver, 'override_existing_value_override')

        logging.info('Beginning import.')
        click_element_by_id(self.driver, 'btnImport')

        logging.info('Checking for message to indicate processing is complete.')
        finished_processing_text = f'Processed {num_rows_in_file} out of {num_rows_in_file} records'
        done_processing = False

        num_of_loops = math.floor(max_processing_wait_time_in_seconds // 10)
        
        for i in range(num_of_loops):
            try:
                logging.info(f'Check #{i + 1} of {num_of_loops}')
                wait_for_element_containing_specific_text(self.driver, finished_processing_text, 10)
                logging.info('File is done processing.')
                done_processing = True
                break
            except:
                logging.info('Message not found yet. Refreshing.')
                self.refresh()

        if done_processing == False:
            raise Exception(f'Something went wrong. File did not finish processing within {max_processing_wait_time_in_seconds} seconds.')

        logging.info('Checking whether all records imported successfully.')
            
        successful_import_text = f'Imported:  {num_rows_in_file}'
        imported_element_text = self.driver.find_element(By.XPATH, '/html/body/div[2]/div[4]/div[2]/div[2]/h3').text 
        # Using the old wait_for_element_containing_specific_text() function did not work for finding the right message, so the above
        #    line gets the "Imported:  X" message 

        assert imported_element_text == successful_import_text, 'Import message indicates not all files imported successfully.'

        logging.info('All records imported successfully. Upload is complete.')

    def _log_into_powerschool_admin(self, username, password):
        """
        Log into PowerSchool Admin and confirm the login was successful.
        
        Parameters:
        n/a
            
        Returns:
        n/a
        """
        
        logging.info("Create webpage URL for PowerSchool Admin")
        host_full = self.host + "/" + ADMIN_LOGIN_PAGE_PATH
        logging.info(f"The webpage url is: {host_full}")
        
        logging.info("Go to webpage URL for PowerSchool Admin")
        self.driver.get(host_full)
        
        logging.info("Find the username field within the login page")
        elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((By.ID, 
            'fieldUsername')))
        
        logging.info("Clear any pre-filled values within the username field")
        elem.clear()
        
        logging.info("Type your PowerSchool username")
        elem.send_keys(username)
        
        logging.info("Find the password field within the HTML page")
        elem = self.driver.find_element(By.ID, 'fieldPassword')
        
        logging.info("Type your PowerSchool password")
        elem.send_keys(password)
        
        logging.info("Press enter to submit your credentials and complete your login.")
        elem.send_keys(Keys.RETURN)

        try:
            logging.info("Waiting for 'Start Page' element to be visible to confirm successful login")
            elem = WebDriverWait(self.driver, 30).until(EC.visibility_of_element_located((By.XPATH, 
                "//h1[text()='Start Page']")))

            logging.info("Successful login confirmed!")
        except:
            raise Exception("Unable to confirm successful login to PowerSchool. Please check your \
                credentials.")

    def _get_current_domain(self):
        """
        Retrieves the current domain.
        
        Parameters:
        self

        Returns:
        str: The current domain
        """

        return self.driver.current_url[8:].split("/",1)[:1][0]

    def _get_current_path(self):
        """
        Retrieves the current path, minus the 'https://'.
        
        Parameters:
        self

        Returns:
        str: The current path
        """

        return self.driver.current_url[8:].split("/",1)[1:][0]

    def _wait_for_new_file_in_folder(self, folder_path, original_files, max_attempts=20000):
        """
        Waits until a new file shows up in a folder. Loops until that is true or max_attempts is 
        reached.

        Parameters:
        self
        folder_path: The folder being monitored.
        original_files: The list of files originally in the folder, before the new one is added.
        max_attmepts: Optional parameter that sets how many loops the function will do.

        Returns:
        n/a
        """
        file_added = False
        attempts = 0
        while True and attempts < max_attempts:
            for root, folders, files in os.walk(folder_path):
                # break 'for' loop if files found
                if len(files) > len(original_files):
                    file_added = True
                    break
                else:
                    continue
            # break 'while' loop if files found
            if file_added:
                # wait for download to complete fully after it's been added - hopefully 3 seconds 
                #    is enough.
                time.sleep(3)
                break
            attempts +=1

    def _rename_recent_file_in_dir(self, folder, append_text):
        """
        Gets the most recent file in a folder and appends text to its filename.

        Parameters:
        self
        folder: The path of the folder
        append_text: The text to appended to the filename before the extension.

        Returns:
        n/a
        """

        recent_file = get_most_recent_file_in_dir(folder)
        recent_file = recent_file.replace('\\', '/')
        new_file, file_ext = os.path.splitext(recent_file)
        new_file += append_text
        new_file += file_ext
        os.rename(recent_file, new_file)


# Code adapted from original code by Yusuph in the dops-calpad repo

# Author: Howard Shen
# Last Edited 5/25/2023

import logging
import os
import pandas as pd
import tempfile
import time

from datetime import date, datetime

from selenium.webdriver.support.ui import Select, WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import (
    TimeoutException,
    NoSuchElementException,
    ElementNotInteractableException
)

from ducttape.utils import (
    DriverBuilder,
    get_most_recent_file_in_dir,
    wait_for_any_file_in_folder
)
    
try:
    from spswarehouse.credentials import calpads_config
except ModuleNotFoundError:
    print("No credentials file found in spswarehouse. This could cause issues.")

from spswarehouse.calpads.calpads_config import (
    monitoring_report_base,
    monitoring_links,
    snapshot_report_base,
    snapshot_links,
)
    
class CALPADS():
    
    def __init__(
        self,
        config=None,
        username=None,
        password=None,
        host=None,
        download_location=None,
        headless=True,
    ):
        """
        By default, the class will pull the username and password from the
        credentials file in the module. You can override the credentials file
        by passing either a config dictionary or both a username and password.
        
        Parameters:
        config: A dictionary containing a username and password key. Optional.
            Supercedes the username and password in credentials.py in this module.
            Also supercedes a username and password combo passed to this function in
            addition to the config.  May also contain a host key that will supercede
            both the host parameter and the host from credentials.py
        username: A CALPADS username, which should be in the form of an email address.
            Must be paired with a password. Supercedes credentials.py.
        password: Password for the given CALPADS username. Must be paired with a username
            Supercedes credentials.py
        host: The URL for CALPADS, in the format `https://www.calpads.org`. Optional.
            Supercedes the host from credentials.py
        download_location: The local folder that you want to save files too. If no
            folder path passed, creates a temporary directory for this object.
        headless: Selenium headless value. Default to True. If using this in a notebook,
            recommend setting to False.
        """
        
        self.host = None
        if config is not None:
            username = config['username']
            password = config['password']
            if 'host' in config:
                self.host = config['host']
        elif username is not None and password is not None:
            username = username
            password = password
        else:
            username = calpads_config['username']
            password = calpads_config['password']
        
        if self.host is None and host is not None:
            self.host = host
        elif self.host is None:
            self.host = calpads_config['host']
        else:
            pass
        
        if download_location is None:
            self.download_location = tempfile.mkdtemp()
        else:
            self.download_location = download_location
    
        self.driver = DriverBuilder().get_driver(
            download_location=self.download_location,
            headless=headless
        )
        self._login_to_calpads(username, password)

    def quit(self):
        self.driver.quit()

    def upload_file(self, lea, submission_type, file_path):
        """
        To be used in conjunction with the create_extract method, uploads the extract via File Submission on CALPADS.

        Parameters:
        lea: The numerical value of the LEA on the CALPADS site. Find by inspecting the
            dropdown on the CALPADS page.
        submission_type: The four letter code, in all caps, for the file type you're 
            uploading
        file_path: The file path to the file you want to upload. Absolute path is
            recommended (`os.path.abspath()`). Relative path behavior untested

        Returns:
        bool: True for a successful upload, else False
        """
        
        self._select_lea(lea)
        self.driver.get(f'{self.host}/FileSubmission/FileUpload')
#         self.driver.refresh()
        
        try:
            elem = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((
                By.XPATH,
                '//*[@id="btnSearch"]'
            )))
        except TimeoutException:
            logging.info('Unable to successfully reach the File Submission web page for {}. Closing the driver.'.format(lea))
            return False

        #Choose file type for the upload
        select = Select(self.driver.find_element(By.XPATH, '//*[@id="tbFileUpload"]/tbody/tr/td[2]/select'))
        select.select_by_value(submission_type)

        #Find the file to upload by pulling the most recent file in the specific illuminate extract folder
        choose_file = self.driver.find_element(By.XPATH, '//*[@id="tbFileUpload"]/tbody/tr/td[4]/input')
        choose_file.send_keys(os.path.abspath(file_path))

        #Upload
        button = self.driver.find_element(By.XPATH, '//*[@id="btnSearch"]')
        button.click()

        try:
            success = WebDriverWait(self.driver, 30).until(EC.presence_of_element_located(
            (By.CSS_SELECTOR, 'div.alert.alert-success.alert-dismissible.fade.in'))) #TODO: Confirm this reliably works. Otherwise, use text.
        except TimeoutException:
            logging.info('Something went wrong with the file upload for {}. Review files in directories and try again.'.format(lea)) #TODO: CALPADS provides error alerts in red on same page, should send to user
            #TODO: Provide more information on error if possible?
            return False
        else:
            logging.info('{} extract successfully uploaded for {}. Check the file submission status later.'.format(type(self).__name__, lea))
            return True
        
    def get_upload_results(self, lea, submission_type, max_wait=5):
        """
        Retrieves the validation errors for the most recent upload of the given
        submission type for the given LEA. Must be used within seven days of the upload
        (default range on CALPADS).

        Parameters:
        lea: The numerical value of the LEA on the CALPADS site. Find by inspecting the
            dropdown on the CALPADS page.
        submission_type: The four letter code, in all caps, for the file type you're 
            uploading
        max_wait: The maximum number of minutes to wait if the submission is still processing.
            Default is 5 minutes. Must be at least 1.

        Returns:
        submitted_date_string (string): ISO formate date of the submission whose errors were checked. Returns None if the file submission can't be found for whatever reason
        job_results_df (DataFrame): Returns a dataframe of the job results or False if the job is not yet in a final status.
        error_details (DataFrame): Returns a dataframe of error details, False if unable to find error details, True if file posted.
        """

        if(max_wait < 1):
            raise Exception('max_wait parameter must be 1 or more.')

        self._select_lea(lea)
        self.driver.get(f'{self.host}/FileSubmission/')
        try:
            file_elem = WebDriverWait(self.driver, 15).until(
                EC.presence_of_element_located((By.ID, 'FileType'))
            )
        except TimeoutException:
            logging.info("Can't load View Submission page")
            return None, False

        # Filter the file list
        file_select = Select(file_elem)
        file_select.select_by_value(submission_type)
        apply_button = self.driver.find_element(By.ID, "fileSubSearch")
        apply_button.click()
        try:
            # The filtered page is identical, so sleep for a moment
            time.sleep(2)
            WebDriverWait(self.driver, 20).until(EC.text_to_be_present_in_element(
                (By.XPATH, '//*[@id="FileSubmissionSearchResults"]/table/tbody/tr[1]/td[6]'),
                submission_type
            ))
        except TimeoutException:
            logging.info(f"No recent submissions of type {submission_type}")
            return None, False

        # Process submission date
        date_text = self.driver.find_element(
            By.XPATH,
            '//*[@id="FileSubmissionSearchResults"]/table/tbody/tr[1]/td[3]'
        ).text

        submitted_datetime = datetime.strptime(date_text, "%m/%d/%Y %I:%M:%S %p")
        submitted_date_string = submitted_datetime.date().isoformat()

        for i in range(max_wait):
            file_status = self.driver.find_element(
                By.XPATH,
                '//*[@id="FileSubmissionSearchResults"]/table/tbody/tr[1]/td[8]'
            ).text
            if file_status in ['Posted', 'Failed', 'Rejected']:
                logging.info('Found final job results. Storing them.')
                job_results_df = pd.read_html(self.driver.page_source)[0].head(1)
                break
            else:
                if((i+1)< max_wait):
                    logging.info(f'Try #{i+1}: Did not find finished job. Will try again in 60 seconds.')
                    time.sleep(60)
                    self.driver.refresh()
                    time.sleep(3)
                else:
                    logging.info(f'Try #{i+1}: Did not find finished job.')

        # Branch based on file status
        if file_status == 'Posted':
            logging.info(f"{submission_type} Posted on {submitted_date_string}")
            return submitted_date_string, job_results_df, True
        elif file_status == 'Failed':
            logging.info(f"{submission_type} Failed on {submitted_date_string}. Check manually")
            return submitted_date_string, job_results_df, False
        elif file_status == 'Rejected':
            view_errors = self.driver.find_element(
                By.XPATH,
                '//*[@id="FileSubmissionSearchResults"]/table/tbody/tr[1]/td[1]/a'
            )
            view_errors.click()
        else:
            logging.info(f"{submitted_date_string} {submission_type} not complete after {max_wait} minute{'s' if max_wait > 1 else ''}.")
            return submitted_date_string, False, False

        try:
            WebDriverWait(self.driver, 30).until(EC.text_to_be_present_in_element((
                By.XPATH,
                '//*[@id="main"]/div/div[2]/header/h1'), 'View Submission Details'
            ))
            time.sleep(10) # The file errors takes a moment to load; extended from 2 to 10 seconds after some rejected files didn't have errors captured
        except TimeoutException as t:
            logging.info('Something went wrong with loading the submission error details page.')
            return submitted_date_string, job_results_df, False

         #We expect the error details to be the second table in the HTML source
        error_details_df = pd.read_html(self.driver.page_source)[1]

        # Get errors if there are multiple pages
        error_has_next = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.XPATH, '//*[@id="FileErrors"]/div/a[3]')))

        all_error_details = []
        all_error_details.append(error_details_df)

        while 'disabled' not in error_has_next.get_attribute('class'):
            logging.info('Next Page')
            error_has_next.click()
            # Give time for next page to load
            time.sleep(10)
            next_details_df = pd.read_html(self.driver.page_source)[1] #TODO: Confirm that this is a completely new df
            all_error_details.append(next_details_df)
            error_has_next = WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.XPATH, '//*[@id="FileErrors"]/div/a[3]')))

        all_error_details = pd.concat(all_error_details)
        logging.info("Successfully collected all errors from file submission error details for the {} {} extract.".format(submitted_date_string, submission_type))
        return submitted_date_string, job_results_df, all_error_details
    
    def get_certification_status(self, lea, academic_year, submission_name, **kwargs):
        """
        Retrieves the status of a Certification window.

        Parameters:
        lea: The numerical value of the LEA on the CALPADS site. Find by inspecting the
            dropdown on the CALPADS page.
        academic_year: Integer representing the academic year that you want to check
            (per team norms, this is the year when the school year ends)
        submission_name: The name of the certification window. As of this edit, the
            certification windows are Fall1, Fall2, EOY1, EOY2, EOY3, and EOY4.
        optional arguments (kwargs):
            rollover_date: Applies only to EOY3. An ISO format string (YYYY-MM-DD)
                representing the expected rollover date of your SIS. The SIS rollover
                is when exit dates and exit codes get put on enrollment records.
                Prior to this, CALPADS will return CERT131 errors for every actively
                enrolled student. This function will begin scraping CERT131 the day
                after the rollover date. If you do not pass a rollover date, CERT131
                will be scraped.

        Returns:
        cert_status (Boolean): Returns False if the certification window isn't open yet;
            returns True if it is.
        error_details (DataFrame): Returns a dataframe of error details, None if no errors,
            None if certification window isn't open.
        warning_details (DataFrame): Returns a dataframe of warningsdetails, None if no
            warnings, None if certification window isn't open.
        """
        
        # Fall1, EOY3, EOY4 have a SELPA approval, which
        # creates an extra table on the Cert status page, altering the XPATH
        if submission_name in ['Fall1', 'EOY3', 'EOY4']:
            data_div_num = 5
            data_table_num = 3
            error_table_num = 1
            warning_table_num = 2
        else:
            data_div_num = 4
            data_table_num = 2
            error_table_num = 0
            warning_table_num = 1
        
        # CERT131 is somewhat meaningless prior to the end of the year
        if submission_name == 'EOY3':
            try:
                rollover_date = date.fromisoformat(kwargs['rollover_date'])
                cert131_flag = date.today() > rollover_date
                logging.info(f"Today is {date.today()}, Rollover date is {rollover_date}, CERT131 = {cert131_flag}")
            except KeyError:
                # KeyError means rollover_date was not passed as an argument
                logging.info("No rollover date passed; scraping CERT131")
                cert131_flag = True
        else:
            cert131_flag = True
        
        self._select_lea(lea)
        
        academic_year_string = f"{academic_year-1}-{academic_year}"
        self.driver.get(f"https://www.calpads.org/StateReporting/Certification?AcademicYear={academic_year_string}&Snapshot={submission_name}")
        
        try:
            WebDriverWait(self.driver, 20).until(EC.presence_of_element_located((
                By.XPATH,
                '/html/body/div[1]/main/div/div[2]/div[3]/div/div/div/div/table/tbody/tr/td[1]'
            )))
        except TimeoutException:
            logging.info(f"{submission_name} certification errors not available yet")
            return False, None, None
        
        error_and_warn_text = self.driver.find_element(
            By.XPATH,
            "/html/body/div[1]/main/div/div[2]/div[3]/div/div/div/div/table/tbody/tr/td[9]"
        ).text
        error_and_warn_count = int(error_and_warn_text)

        if error_and_warn_count > 0:
            cert_status_link = self.driver.find_element(
                By.XPATH,
                '/html/body/div[1]/main/div/div[2]/div[3]/div/div/div/div/table/tbody/tr/td[1]/a'
            )
            cert_status_link.click()
            try:
                WebDriverWait(self.driver, 20).until(EC.text_to_be_present_in_element(
                    (
                        By.XPATH,
                        '/html/body/div/main/div/div[2]/header/h1'
                    ),
                    'Certification Details - LEA'
                ))
            except TimeoutException:
                logging.info("Failed to load Certification Status page")
                raise RuntimeError("Failed to load Certification Status page")
            try:
                WebDriverWait(self.driver, 15).until(EC.presence_of_element_located((
                    By.XPATH,
                    f'/html/body/div/main/div/div[2]/div[{data_div_num}]/div[1]/div[1]/div/ul/li/div/div/div/table/tbody/tr/td[4]/a'
                )))
                # Error list should be the second table
                errors_table = pd.read_html(self.driver.page_source)[error_table_num]
                error_df = pd.DataFrame()
                
                error_list_has_next = self.driver.find_element(
                    By.XPATH,
                    '//*[@id="RequiredGrid_FATAL"]/div/a[3]'
                )
                while 'disabled' not in error_list_has_next.get_attribute('class'):
                    logging.info("Getting next page of errors list")
                    try:
                        error_list_has_next.click()
                    except ElementNotInteractableException:
                        banner = self.driver.find_element(
                            By.XPATH,
                            '//*[@id="CertificationErrorSummary"]/li/a'
                        )
                        banner.click()
                        error_list_has_next.click()
                    time.sleep(2)
                    errors_table = pd.concat(
                        [errors_table, pd.read_html(self.driver.page_source)[error_table_num]],
                        axis=0,
                        ignore_index=True
                    )
                    error_list_has_next = self.driver.find_element(
                        By.XPATH,
                        '//*[@id="RequiredGrid_FATAL"]/div/a[3]'
                    )
                
                for i in range(len(errors_table)):
                    error_id = errors_table["Message ID"][i]

                    if error_id == 'CERT131' and not cert131_flag:
                        logging.info("Skipping CERT131")
                        continue
                        
                    logging.info(f"Getting list for {error_id}")
                    error_dropdown = self.driver.find_element(
                        By.ID,
                        'CertificationEditQuery_ChildErrorCategory'
                    )
                    error_select = Select(error_dropdown)
                    error_select.select_by_value(error_id)
                    apply_button = self.driver.find_element(
                        By.XPATH,
                        '//*[@id="certDetails"]/div[1]/div[3]/div/div/form/div[3]/button'
                    )
                    apply_button.click()
                    
                    try:
                        WebDriverWait(self.driver, 30).until(EC.text_to_be_present_in_element(
                            (
                                By.XPATH,
                                f'/html/body/div/main/div/div[2]/div[{data_div_num}]/div[1]/div[4]/div/div/table/tbody/tr/td[10]/a'
                            ),
                            error_id
                        ))
                    except TimeoutException:
                        logging.info(f"Can't load errors for {error_id}")
                        raise RuntimeError(f"Can't load errors for {error_id}")
                    
                    error_has_next = self.driver.find_element(
                        By.XPATH,
                        '//*[@id="RequiredGrid"]/div/a[3]'
                    )
                    
                    error_df = pd.concat(
                        [error_df, pd.read_html(self.driver.page_source)[data_table_num]],
                        axis=0,
                        ignore_index=True,
                    )
                    
                    while 'disabled' not in error_has_next.get_attribute('class'):
                        logging.info("Getting next page")
                        error_has_next.click()
                        time.sleep(2)
                        error_df = pd.concat(
                            [error_df, pd.read_html(self.driver.page_source)[data_table_num]],
                            axis=0,
                            ignore_index=True,
                        )
                        error_has_next = self.driver.find_element(
                        By.XPATH,
                        '//*[@id="RequiredGrid"]/div/a[3]'
                        )
                    
                # Drop the "View" button column
                if len(error_df) == 0:
                    logging.info("Only Cert Error was CERT131, and rollover has not passed.")
                    error_df = None
                else:
                    error_df.drop(columns="Error Record", inplace=True)
                
            except TimeoutException:
                logging.info(f"No errors for {submission_name}")
                error_df = None
            
            # Get Warnings
            try:
                WebDriverWait(self.driver, 5).until(EC.presence_of_element_located((
                    By.XPATH,
                    f'/html/body/div/main/div/div[2]/div[{data_div_num}]/div[1]/div[2]/div/ul/li/div/div/div/table/tbody/tr/td[4]/a'
                )))
                # Warning list should be the third table
                warnings_table = pd.read_html(self.driver.page_source)[warning_table_num]
                warning_df = pd.DataFrame()
                
                warning_list_has_next = self.driver.find_element(
                    By.XPATH,
                    '//*[@id="RequiredGrid_WARN"]/div/a[3]'
                )
                while 'disabled' not in warning_list_has_next.get_attribute('class'):
                    logging.info("Getting next page of warnings list")
                    try:
                        warning_list_has_next.click()
                    except ElementNotInteractableException:
                        banner = self.driver.find_element(
                            By.XPATH,
                            '//*[@id="CertificationWarnSummary"]/li/a'
                        )
                        banner.click()
                        warning_list_has_next.click()
                    time.sleep(2)
                    warnings_table = pd.concat(
                        [warnings_table, pd.read_html(self.driver.page_source)[warning_table_num]],
                        axis=0,
                        ignore_index=True
                    )
                    warning_list_has_next = self.driver.find_element(
                        By.XPATH,
                        '//*[@id="RequiredGrid_WARN"]/div/a[3]'
                    )
                
                for i in range(len(warnings_table)):
                    warning_id = warnings_table["Message ID"][i]
                        
                    logging.info(f"Getting list for {warning_id}")
                    warning_dropdown = self.driver.find_element(
                        By.ID,
                        'CertificationEditQuery_ChildErrorCategory'
                    )
                    warning_select = Select(warning_dropdown)
                    warning_select.select_by_value(warning_id)
                    apply_button = self.driver.find_element(
                        By.XPATH,
                        '//*[@id="certDetails"]/div[1]/div[3]/div/div/form/div[3]/button'
                    )
                    apply_button.click()
                    
                    try:
                        WebDriverWait(self.driver, 15).until(EC.text_to_be_present_in_element(
                            (
                                By.XPATH,
                               f'/html/body/div/main/div/div[2]/div[{data_div_num}]/div[1]/div[4]/div/div/table/tbody/tr/td[10]/a'
                            ),
                            warning_id
                        ))
                    except TimeoutException:
                        logging.info(f"Can't load errors for {warning_id}")
                        raise RuntimeError(f"Can't load errors for {warning_id}")
                    
                    warning_has_next = self.driver.find_element(
                        By.XPATH,
                        '//*[@id="RequiredGrid"]/div/a[3]'
                    )
                    
                    warning_df = pd.concat(
                        [warning_df, pd.read_html(self.driver.page_source)[data_table_num]],
                        axis=0,
                        ignore_index=True,
                    )
                    
                    while 'disabled' not in warning_has_next.get_attribute('class'):
                        logging.info("Getting next page")
                        warning_has_next.click()
                        time.sleep(2)
                        warning_df = pd.concat(
                            [warning_df, pd.read_html(self.driver.page_source)[data_table_num]],
                            axis=0,
                            ignore_index=True,
                        )
                        warning_has_next = self.driver.find_element(
                            By.XPATH,
                            '//*[@id="RequiredGrid"]/div/a[3]'
                        )
                    
                # Drop the "View" button column
                warning_df.drop(columns="Error Record", inplace=True)
                
            except TimeoutException:
                logging.info(f"No warnings for {submission_name}")
                warning_df = None
                
            return True, error_df, warning_df
        elif error_and_warn_count == 0:
            logging.info(f"No errors or warnings for {submission_name}")
            return True, None, None
        else:
            raise RuntimeError("Impossible part of error_and_warn_count if-elif-else reached.")

    def download_monitoring_report(
        self,
        lea,
        academic_year,
        report_code,
        report_date=None,
        download_type='csv',
        max_wait_time=10
    ):
        """
        Download a CALPADS snapshot report.
        
        Known issue: The download folder needs to be empty when this function is called.
        
        Parameters:
        lea: The numerical value of the LEA on the CALPADS site. Find by inspecting the
            dropdown on the CALPADS page.
        academic_year: Integer representing the academic year that you want to check
            (per team norms, this is the year when the school year ends)
        submission_name: The name of the certification window. As of this edit, the
            certification windows are Fall1, Fall2, EOY1, EOY2, EOY3, and EOY4.
        report_code: The code for the report that you want. Should be a string in
            the form of "#.#". Check calpads_config.py for the list of available
            report_code + submission_name combos
        report_date (str): An ISO format date string ('YYYY-MM-DD') indicating the 
            desired date for the report. If None, the function will download whatever
            the default date on the report is (usually today).
            Note that the function only uses the month and day. (E.g., if you pass
            academic_year=2023 and report_date="2024-10-01", the report downloaded will be
            for October 1 of academic year 2023, which is actually "2022-10-01")
        download_type (str): The format in which you want the download for the report.
            Currently supports csv, excel, and pdf.
        max_wait_time: Integer >=0 indicating the maximum number of minutes to wait
            for the report to generate and for the download to succeed. This means
            this function can actually take up to 2*max_wait_time to run.
        
        Returns:
        string: The filepath to the downloaded file
        """
        url_tail = monitoring_links[report_code]
        report_url = self.host + monitoring_report_base + url_tail
        
        return self._download_report(
            lea,
            report_url,
            academic_year,
            download_type,
            max_wait_time,
            report_date=report_date
        )
            
    def download_snapshot_report(
        self,
        lea,
        academic_year,
        submission_name,
        report_code,
        cert_status=None,
        download_type='csv',
        max_wait_time=10
    ):
        """
        Download a CALPADS snapshot report.
        
        Known issue: The download folder needs to be empty when this function is called.
        
        Parameters:
        lea: The numerical value of the LEA on the CALPADS site. Find by inspecting the
            dropdown on the CALPADS page.
        academic_year: Integer representing the academic year that you want to check
            (per team norms, this is the year when the school year ends)
        submission_name: The name of the certification window. As of this edit, the
            certification windows are Fall1, Fall2, EOY1, EOY2, EOY3, and EOY4.
        report_code: The code for the report that you want. Should be a string in
            the form of "#.#". Check calpads_config.py for the list of available
            report_code + submission_name combos
        cert_status: If there are multiple snapshots, which version you want. If
            `None`, the top most option is selected (usually some variant of
            "Certified" if the snapshot is certified). Will raise an error if the
            given cert_status is not available.
        download_type (str): The format in which you want the download for the report.
            Currently supports csv, excel, and pdf.
        max_wait_time: Integer >=0 indicating the maximum number of minutes to wait
            for the report to generate and for the download to succeed. This means
            this function can actually take up to 2*max_wait_time to run.
        
        Returns:
        string: The filepath to the downloaded file
        """
        url_tail = snapshot_links[submission_name][report_code]
        report_url = self.host + snapshot_report_base + url_tail
        
        return self._download_report(
            lea,
            report_url,
            academic_year,
            download_type,
            max_wait_time,
            cert_status=cert_status
        )
            
    def _login_to_calpads(self, username, password):
        self.driver.get(self.host)
        try:
            WebDriverWait(self.driver, 15).until(EC.presence_of_element_located((
                By.XPATH,
                '/html/body/div[3]/div/form/div/div[2]/fieldset/div[4]/div/button'
            )))
        except TimeoutException:
            logging.info("Was unable to reach the login page. Check the browser: {}".format(self.driver.title))
            return False
        except NoSuchElementException:
            logging.info("Was unable to reach the login page. Check the browser: {}".format(self.driver.title))
            return False
        user = self.driver.find_element(By.ID, "Username")
        user.send_keys(username)
        pw = self.driver.find_element(By.ID, "Password")
        pw.send_keys(password)
        agreement = self.driver.find_element(By.ID, "AgreementConfirmed")
        self.driver.execute_script("arguments[0].click();", agreement)
        btn = self.driver.find_element(
            By.XPATH,
            "/html/body/div[3]/div/form/div/div[2]/fieldset/div[4]/div/button"
        )
        btn.click()
        try:
            WebDriverWait(self.driver, 30).until(EC.presence_of_element_located((
                By.ID,
                'org-select'
            )))
        except TimeoutException:
            logging.info('Something went wrong with the login. Checking to see if there was an expected error message.')
            try:
                alert = WebDriverWait(self.driver, 15).until(EC.presence_of_element_located((
                    By.XPATH,
                    '/html/body/div[3]/div/form/div[1]'
                )))
                if 'alert' in alert.get_attribute('class'):
                    logging.info("Found an expected alert during login: '{}'".format(
                        self.driver.find_element(
                            By.XPATH, '/html/body/div[3]/div/form/div[1]/div/ul/li'
                        ).text
                    ))
                    return False
                else:
                    logging.info('There was an unexpected message during login. See driver.')
                    return False
            except TimeoutException:
                logging.info('There was an unexpected error during login. See driver.')
                return False

        return True
    
    def _select_lea(self, lea):
        """
        Factored out common process for switching to a different LEA in the dropdown

        Parameters:
        lea: The numerical value of the LEA on the CALPADS site. Find by inspecting the
            dropdown on the CALPADS page.
        """
        select = Select(self.driver.find_element(By.ID, 'org-select'))
        select.select_by_value(lea)
        WebDriverWait(self.driver, 30).until(EC.element_to_be_clickable((By.ID, 'org-select')))

    def _download_report(
        self,
        lea,
        url,
        academic_year,
        download_type,
        max_wait_time,
        **kwargs
    ):
        """
        Helper function for downloading from the reports iframe
        Will crash if you pass "cert_status" in kwargs but the report is not a Snapshot
            report
        Will also crash if you pass a value othere than None for "report_date" in kwargs
            but the report is not an Accountability/Monitoring report
        """
        
        self._select_lea(lea)
        self.driver.get(url)
        try:
            # The Reports module is in an iframe
            iframe = WebDriverWait(self.driver, 30).until(
                EC.presence_of_element_located((
                    By.XPATH,
                    '//*[@id="reports"]/div/div/div/iframe'
                ))
            )
        except TimeoutException:
            logging.info("Failed to load report page.")
            raise
        
        self.driver.switch_to.frame(iframe)
        self._select_report_academic_year(academic_year)
        
        if "cert_status" in kwargs:
            cert_status = kwargs["cert_status"]
            cert_status_select = Select(self.driver.find_element(
                By.XPATH,
                '//*[@id="ReportViewer1_ctl08_ctl07_ddValue"]'
            ))
            if cert_status is None:
                cert_status_select.select_by_value("1")
            else:
                cert_status.select_by_visible_text(cert_status)
        
        elif "report_date" in kwargs:
            report_date = kwargs["report_date"]
            if report_date is None:
                pass
            else:
                # You need the month and day as strings without leading zeroes
                month = str(int(report_date[5:7]))
                day = str(int(report_date[8:10]))
                
                month_select = Select(self.driver.find_element(
                    By.XPATH,
                    '//*[@id="ReportViewer1_ctl08_ctl05_ddValue"]'
                ))
                month_select.select_by_value(month)
                self._wait_for_view_report_clickable()
                
                day_select = Select(self.driver.find_element(
                    By.XPATH,
                    '//*[@id="ReportViewer1_ctl08_ctl07_ddValue"]'
                ))
                day_select.select_by_value(day)
                
        submit_button = self._wait_for_view_report_clickable()
        submit_button.click()
        
        self._wait_for_view_report_clickable(max_wait_time)
        filepath = self._download_loaded_report(download_type, max_wait_time)
        return filepath
        
    def _select_report_academic_year(self, academic_year):
        """
        Select the given academic_year for the the reports module. Assumes the driver
        is already clicked into the iframe. This function waits for the "View Report"
        button to be clickable before returning.
        """
        academic_year_string = f"{academic_year-1}-{academic_year}"
        yr = Select(self.driver.find_element(
            By.XPATH,
            '//*[@id="ReportViewer1_ctl08_ctl03_ddValue"]'
        ))
        yr.select_by_visible_text(academic_year_string)
        self._wait_for_view_report_clickable()
        
        
    def _wait_for_view_report_clickable(self, max_attempts=3):
        """
        Check for the delay before webpage allows another change in value
        for the report request
        """
        try:
            max_attempts = int(max_attempts)
        except:
            max_attempts = 1
        for attempt in range(max_attempts):
            try:
                view_report = WebDriverWait(self.driver, 60).until(
                    EC.element_to_be_clickable((
                        By.XPATH,
                        '//*[@id="ReportViewer1_ctl08_ctl00"]'
                    ))
                )
                return view_report
            except TimeoutException:
                logging.info('The Report button has not loaded after 1 minute. Attempt: {}'.format(attempt+1))
        logging.info('Max number of attempts waiting for View Report to be clickable reached and all failed.')
        raise TimeoutException
        
    def _download_loaded_report(self, download_type='csv', max_wait_time=3):
        """
        Downloads the report that is already loaded on the page. Assumes the driver
        is still clicked into the iframe with the report controls.
        
        Return:
        string: The filepath to the downloaded file.
        """
        dl_types = {
            'csv': '//*[@id="ReportViewer1_ctl09_ctl04_ctl00_Menu"]/div[7]/a',
            'pdf': '//*[@id="ReportViewer1_ctl09_ctl04_ctl00_Menu"]/div[4]/a',
            'excel': '//*[@id="ReportViewer1_ctl09_ctl04_ctl00_Menu"]/div[2]/a'
        }
        
        dropdown_btn = self.driver.find_element(
            By.XPATH,
            '//*[@id="ReportViewer1_ctl09_ctl04_ctl00"]'
        )
        dropdown_btn.click()
        try:
            dl_button = WebDriverWait(self.driver, 10).until(
                EC.visibility_of_element_located((
                    By.XPATH,
                    dl_types[download_type]
                ))
            )
        except TimeoutException:
            logging.info("Dropdown menu for download not loading")
            raise
        dl_button.send_keys(Keys.ENTER)
        if wait_for_any_file_in_folder(self.download_location, timeout=max_wait_time*60):
            file_path = get_most_recent_file_in_dir(self.download_location)
            logging.info(f"File found: {file_path}")
            return file_path
        else:
            logging.info("No file found")
            return None


monitoring_report_base = "/Report/Accountability/"
monitoring_links = {
    "15.1": "15_1_-_Cohort_Outcome_-_Counts_and_Rates",
    "15.2": "15_2_-_Cohort_Outcome_-_Student_Details",
    "16.7": "16_7_-_Students_with_Disabilities_-_Monitoring_Counts_ODS",
    "16.8": "16_8_-_Students_with_Disabilities_-_Monitoring_Student_List_ODS",
}

snapshot_report_base = "/Report/Snapshot/"
snapshot_links = {
    "Fall1": {
        "1.1": "1_1_EnrollmentPrimaryStatusbySubgroup",
        "1.2": "1_2_EnrollmentPrimaryStatusStudentList",
        "1.3": "1_3_EnrollmentPrimaryStatusDisaggregated",
        "1.4": "1_5_EnrollmentByStatusDisaggregated",
        "1.5": "1_5_EnrollmentByStatusDisaggregated",
        "1.8": "1_8_DropoutsBySubgroupStudentList",
        "1.12": "1_12_DropoutsDisaggregated",
        "1.13": "1_13_ExitCountsDisaggregated",
        "1.14": "1_14_DropoutCountsStateView",
        "1.17": "1_17_FRPM_EnglishLearnerFosterYouthCount_AY23",
        "1.18": "1_18_FPRM_EnglishLearnerFosterYouthStudentList_AY23",
        "1.24": "1_24_DropoutsBySubgroup",
        "2.1": "2_1_TitleIIIEligibleImmigrantCounts",
        "2.2": "2_2_TitleIIIEligibleImmigrantsCountsByBirthCountry",
        "2.8": "2_8_ELAStatusCountByPrimaryLanguage_Fall_",
        "2.9": "2_9_ELASCensusComparison_Fall1_",
        "2.10": "2_10_EnglishLanguageAcquisitionStatusCensusComparisonStudentList",
        "2.14": "2_14_LanguageInstructionProgramCount",
        "2.15": "2_15_LanguageInstruction_StudentList",
        "8.1": "8_1_StudentProfileList",
        "8.1a": "8_1a_StudentProfileExitsList",
        "8.1b": "8_1b_StudentProfileDropoutsList",
        "8.1c": "8_1c_StudentProfileDropouts_StateView",
        "16.1": "16_1_SPEDEducationPlanByPrimaryDisabilityCount",
        "16.2": "16_2_SPEDPrimaryDisabilityByFederalSettingCount",
        "16.3": "16_3_SPEDStudentProfileList",
        "16.5": "16_5_StudentsWithDisabilitiesStudentServices",
        "16.6": "16_6_StudentsWithDisabilitiesStudentServicesStudentList",
        "16.11": "16_11_SPEDPrimaryDisabilityAnnualComparisionReport",
    },
    "Fall2": {
        "2.4": "2_4_EnglishLearnerEducationServicesStudentCountsUnduplicated",
        "2.5": "2_5_EnglishLearnerEducationServicesUnduplicatedCountOfTeachersProvidingELServices",
        "2.6": "2_6_EnglishLearnerCountByPrimaryLanguage",
        "2.7": "2_7_EnglishLearnerEducationServicesStudentList",
        "2.11": "2_11_EnglishLearnerInstructionalPlanCountByPlan",
        "3.1": "3_1_ClassSizeByContentArea",
        "3.2": "3_2_ClassEnrollmentByContentArea",
        "3.3": "3_3_ClassEnrollmentStudentList",
        "3.6": "3_6_CourseSectionEnrollmentCountsByContentArea",
        "3.7": "3_7_CourseSectionEnrollmentCountAndDetails",
        "3.8": "3_8_CourseSectionEnrollmentStudentList",
        "4.1": "4_1_StaffCountFTEByJobClassification",
        "4.2": "4_2_StaffCountAndFTEbyJobClassificationDisaggregated",
        "4.3": "4_3_StaffTeachingAssignmentsDetail",
        "4.4": "4_4_StaffProfileList",
        "4.5": "4_5_StaffNonClassroomBasedorSupportAssignmentDetail",
        "17.1": "17_1_PostsecondaryStatusOutcomeCount",
        "17.2": "17_2_PostsecondaryStatusOutcomeStudentList",
    },
    "EOY1": {
        "3.9": "3_9_CourseSectionsCompletedCountByContentAreaForDepartmentalizedCourses",
        "3.10": "3_10_CourseSectionsCompletedCountAndDetailsForDepartmentalizedCourses",
        "3.11": "3_11_CourseSectionsCompletedStudentListForDepartmentalizedCourses",
        "3.14": "3_14_CareerTechnicalEducationParticipantsCount",
        "3.15": "3_15_CareerTechnicalEducationParticipantsStudentList",
        "3.16": "3_16_EducationalOptionsCourseCompletionStudentCount",
        "3.19": "3_19_CareerTechnicalEducationCompletersCountbyPathway",
        "3.20": "3_20_CareerTechnicalEducationCompletersStudentList",
        "18.1": "18_1_WorkBasedLearningCount",
        "18.2": "18_2_WorkBasedLearningStudentList",
    },
    "EOY2": {
        "5.1": "5_1_ProgramParticipantsCount",
        "5.2": "5_2_ProgramParticipantsNCLBTitleIPartABasicTargetedEducationServices",
        "5.3": "5_3_ProgramParticipantsStudentList",
    },
    "EOY3": {
        "1.21": "1_21_CumulativeEnrollmentCount",
        "1.22": "1_22_GraduatesandCompletersCount",
        "1.23": "1_23_GraduatesandCompletersStudentList",
        "2.16": "2_16_EnglishLanguageAcquisitionStatusELsReclassifiedRFEP_EOY3",
        "2.17": "2_17_EnglishLanguageAcquisitionStatusELsReclassifiedRFEPStudentList_EOY3",
        "5.4": "5_4_HomelessStudentsEnrolledUnduplicatedCountbySchool",
        "5.5": "5_5_HomelessStudentList",
        "7.10": "7_10_IncidentCount",
        "7.11": "7_11_IncidentResultCount",
        "7.12": "7_12_IncidentResultsStudentList",
        "7.13": "7_13_StudentOffense_CountByOffense",
        "7.14": "7_14_IncidentOffenseStudentList",
        "7.15": "7_15_IncidentActionsPersistentlyDangerousOffenseExpulsions",
        "7.16": "7_16_IncidentRemovalsForStudentsWithDisabilitiesCount",
        "7.17": "7_17_UnilateralRemovalsForStudentsWithDisabilitiesCount",
        "7.18": "7_18_IncidentRemovalsForStudentsWithDisabilitiesStudentList",
        "8.1": "8_1_StudentProfileList_EOY3_",
        "14.1": "14_1_StudentAbsenteeismCount",
        "14.2": "14_2_StudentAbsencesStudentList",
    },
    "EOY4": {
        "16.1": "16_1_SPEDEducationPlanByPrimaryDisabilityCount_EOY4",
        "16.2": "16_2_SPEDPrimaryDisabilityByFederalSettingCount_EOY4",
        "16.3": "16_3_SPEDStudentProfileList_EOY4",
        "16.5": "16_5_StudentsWithDisabilitiesStudentServices_EOY4",
        "16.6": "16_6_StudentsWithDisabilitiesStudentServicesStudentList_EOY4",
        "16.9": "16_9_SPEDProgramExitCountbyPrimaryDisability_EOY4",
        "16.10": "16_10_SPEDProgramExitStudentList_EOY4",
        "16.11": "16_11_SPEDPrimaryDisabilityAnnualComparisionReport_EOY4",
        "17.3": "17_3_PostsecondarySurveyOutcomeforSWDsCount",
        "17.4": "17_4_PostsecondarySurveyOutcomeforSWDsStudentList",
    },
}
    


from datetime import datetime, date
import time
import pandas as pd

from .powerschool import PowerSchool

from spswarehouse.general.selenium import (
    helper_type_in_element_by_name,
    helper_select_visible_text_in_element_by_name,
    helper_click_element_by_id,
    helper_ensure_checkbox_is_unchecked_by_name,
    helper_ensure_checkbox_is_checked_by_name,
    helper_type_in_element_by_id,
    helper_select_visible_text_in_element_by_id,
    ensure_checkbox_is_checked_by_name,
    type_in_element_by_name,
    select_visible_text_in_element_by_name,
    click_element_by_id,
)

# Used for making the standard modifications to the Student English Language Acquistion (SELA) upload file; column #'s from the CALPADS file specifications
SELA_COLUMN_NAMES = [
    'Record Type Code', # 12.01
    'Transaction Type Code', # 12.02
    'Local Record ID', #12.03
    'Reporting LEA', #12.04
    'School of Attendance', #12.05
    'Academic Year ID', #12.06
    'SSID', #12.07
    'Student Legal First Name', #12.08
    'Student Legal Last Name', #12.09
    'Student Birth Date', #12.10
    'Student Gender Code', #12.11
    'Local Student ID', #12.12
    'English Language Acquisition Status Code', #12.13
    'English Language Acquisition Status Start Date', #12.14
    'Primary Language Code', #12.15
    'Correction Reason Code', #12.16
]

class PowerSchoolCALPADS(PowerSchool):
    """
    This class extends the PowerSchool class in order to download CALPADS
    reports.
    """
    
    def __init__(
        self,
        config: dict=None,
        username: str=None,
        password: str=None,
        host: str=None,
        headless: bool=True,
        download_location: str='.',
        chrome_option_prefs: dict=None,
    ):
        super().__init__(config, username, password, host, headless, download_location, chrome_option_prefs)    
        
    def download_calpads_report_for_school(self, school_full_name: str, submission_window: str, 
        calpads_report_abbreviation: str, ps_school_subdistrict_name: str, file_postfix: str, 
        destination_directory_path: str, report_parameters: dict, ps_school_year_dropdown=None, 
        validation_only_run: bool=False):
        """
        Switches to the desired school in PowerSchool and calls the function to generate the desired
        report for the specified submission window. Note: This function currently only supports EOY 
        reports and some All Year reports. Fall 1 and Fall 2 reports should not use this function until 
        it is expanded.
        """

        if calpads_report_abbreviation not in self.CALPADS_REPORT_TYPES:
            raise KeyError(f"{calpads_report_abbreviation} not found in report types")
            
        report_kwargs = {
            'ps_report_link_text': self.CALPADS_REPORT_TYPES[calpads_report_abbreviation]['ps_title'], 
            'file_postfix': file_postfix, 
            'destination_directory_path': destination_directory_path, 
            'report_parameters': report_parameters,
            'validation_only_run': validation_only_run
        }
        
        # Choose the school year if one is provided
        if ps_school_year_dropdown is not None:
            self.switch_to_school_year(ps_school_year_dropdown)

        # The SCSC/SCSE reports need to be run from the District Office level in order to properly generate 
        #   LEA IDs without dropping leading zeros
        #   SCSC also requires an additional parameter
        if calpads_report_abbreviation in ["SCSC", "SCSE"]:
            self.switch_to_school('District Office')
            report_kwargs['ps_school_subdistrict_name'] = ps_school_subdistrict_name
        else:
            self.switch_to_school(school_full_name)

        # Must pass self to the function explicitly
        # Howard's note: I'm not certain the cause, but I think it's because the functions in the
        # dictionary are not (and cannot be) defined as `self.function`, thus requiring that self
        # be passed explicitly instead
        return self.CALPADS_REPORT_TYPES[calpads_report_abbreviation]['function'](self, **report_kwargs)
    
    # All Year Reports #################

    def _download_all_year_report_for_ssid_enrollment_records_senr(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the SSID Enrollment Records (SENR) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_type_in_element_by_name(self.driver, 'StartDate', 
            report_parameters['report_start_date'])
        helper_type_in_element_by_name(self.driver, 'EndDate', 
            report_parameters['report_end_date'])
        helper_select_visible_text_in_element_by_name(self.driver, 'submissionMode', 
            'Non-submission mode (all records)') # Only 'Non-submission mode' is supported by this tool
        time.sleep(1) # Give page time to react
        helper_select_visible_text_in_element_by_name(self.driver, 'ssidOption', 
            report_parameters['student_selection_filter'])
        helper_select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')

        # Submit report
        helper_click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)
    
    def _download_all_year_report_for_student_english_language_acquisition_records_sela(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Student English Language Acquisition (SELA) report in 
        PowerSchool and downloads it.

        IMPORTANT NOTE #1: You need to run the remove_sela_records_beginning_before_report_start_date()
        function on the final text file to filter out any records whose start
        dates are before your intended report start date. PowerSchool's
        filtering does not work properly for this.

        IMPORTANT NOTE #2: There is an edge case that this code and PowerSchool 
        are not equipped to handle. If you have a student who received an English
        Language Acquisition status at another district before your report_start_date
        and you store that status in PowerSchool, there is no easy way to decline to
        report that record to CALPADS. But if you do report that record, CALPADS will
        reject it and your whole upload file. 
        
        For instance, if a student enrolled at another district and received a TBD 
        status in September and then they joined your district in October, if you 
        enter that TBD status into PowerSchool, the SELA file will error out in CALPADS.
        
        The proper fix is to only include records that start when a student is enrolled
        with you, but that would be complicated to do through the PowerSchool UI. Summit
        will likely solve this issue by manually setting any impacted English Language
        Acquisition statuses to have a start date in PowerSchool of before the school year 
        begins. That way, they'll be excluded from this tool. If this is a problem for you, 
        you may want to do something similar or find a way to pull enrollment data into 
        your code to do additional filtering on the upload file.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)

        # SELA requires some very specific setup to export in a way that CALPADS will accept. Specifically,
        #   you can only include students that are currently rostered to you in CALPADS, so this function
        #   will only pull currently enrolled students or students who were enrolled on the report end date
        #   if that date is in the past. Usually, the report end date will be the last day of school, so this
        #   will enable submissions after the school year has ended.

        report_end_date_datetime_object = datetime.strptime(report_parameters['report_end_date'], '%m/%d/%Y').date()
        today_datetime_object = date.today()

        if today_datetime_object > report_end_date_datetime_object:
            date_for_report = report_parameters['report_end_date']
        else:
            date_for_report = today_datetime_object.strftime('%m/%d/%Y')
        
        # Enter specific parameters for this report
        helper_type_in_element_by_name(self.driver, 'startDate', date_for_report)
        helper_type_in_element_by_name(self.driver, 'endDate', date_for_report)

        helper_type_in_element_by_name(self.driver, 'status_start_date', '') # Make sure this field is cleared
        helper_type_in_element_by_name(self.driver, 'status_end_date', '') # Make sure this field is cleared

        helper_select_visible_text_in_element_by_name(self.driver, 'deltaOff', # Unclear why this is the element name
            'Non-submission mode (all records)') # Only 'Non-submission mode' is supported by this tool
        time.sleep(1) # Give page time to react

        helper_select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        
        # Below defaults to "No Group Selected" because school should already be chosen
        helper_select_visible_text_in_element_by_name(self.driver, 'schoolGroup', '[No Group Selected]') 

        # Submit report
        helper_click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)

    def _download_all_year_report_for_student_information_records_sinf(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Student Information Records (SINF) report in PowerSchool and downloads it.
        This function assumes that the export will be run repeatedly over the course of the year
        and therefore doesn't use the "Check for Fall Submission" functionality. This is because
        there are occasionally needs for updating SINF outside of Fall 1, such as for Pandemic
        EBT benefits.

        Uploading an updated SINF file does not seem to close out existing Demographics and
        Address  records, except when the "Effective Start Date" is greater than the most 
        recent "Effective Start Date" in CALPADS. Thus, the "Make Effective Start Date Match 
        Enrollment Start Date" checkbox (described just below) should mean existing records
        just get updated, rather than new records being created.

        Important notes about how the SINF report parameters in PowerSchool work:
        - The "Check for Fall Submission" checkbox means it will only submit records for students
        enrolled on the Census Date that you enter. This function intentionally does not use
        this setting, so it will guarantee that box is not checked.
        - The "Make Effective Start Date Match Enrollment Start Date" will give all records the 
        student's enrollment date for the year as the start date.
        - The "Start Date" will be the date at which the report looks for enrollments, and it will
        report data for any student enrolled any time up to the "End Date". The "Effective Start
        Date" in the final file will be the student's enrollment date for the year, so long as the
        "Make Effective Start Date Match Enrollment Start Date" box is checked.
        - The "End Date" cannot be greater than the current date for this report, so this function
        will enter the report_end_date or the current date, whichever is earlier.
        - The "Include Students' Preferred Names (If Different From Legal)" will include data in
        those fields if "Yes" is selected. For student privacy reasons, defaulting to "No" here is 
        preferred, especially for students whose lived name may differ from their legal name.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_ensure_checkbox_is_unchecked_by_name(self.driver, 'SubmissionType')
        time.sleep(1) # Give page time to react

        helper_ensure_checkbox_is_checked_by_name(self.driver, 'effectiveStartDate')

        helper_type_in_element_by_name(self.driver, 'StartDate', 
            report_parameters['report_start_date'])
        
        # If the provided report_end_date is in the future, use today's date instead
        end_date_string = report_parameters['report_end_date']
        report_end_date_object = datetime.strptime(end_date_string, '%m/%d/%Y').date()
        current_date = date.today()
        if report_end_date_object > current_date:
            end_date_string = current_date.strftime("%m/%d/%Y")
        helper_type_in_element_by_name(self.driver, 'EndDate', end_date_string)

        helper_select_visible_text_in_element_by_name(self.driver, 'deltaOff',
            'Non-submission mode (all records)') # Only 'Non-submission mode' is supported by this tool
        time.sleep(1) # Give page time to react

        helper_select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        
        # See note in function definition
        helper_select_visible_text_in_element_by_name(self.driver, 'includePreferredName', 'No')

        # Submit report
        helper_click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)


    # Fall 2 Reports ######################

    def _download_fall_2_report_for_staff_demographics_records_sdem(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Staff Demographics Records (SDEM) report in PowerSchool and downloads it.
        Note: This function assumes that the report will be run with data as of Census Day and
        not using the more general Start Date and End Date parameters.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Check the box to indicate this is a Fall submission
        ensure_checkbox_is_checked_by_name(self.driver, 'SubmissionType')
        
        # Date field needs time to appear
        time.sleep(1) 
        type_in_element_by_name(self.driver, 'CensusDate', report_parameters['census_date'])
        
        select_visible_text_in_element_by_name(self.driver, 'deltaOff', 'Non-submission mode (all records)')
        select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        select_visible_text_in_element_by_name(self.driver, 'schoolGroup', 
            '[No Group Selected]') 

        # Submit report
        click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)

    def _download_fall_2_report_for_staff_assignment_records_sass(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Staff Assignment Records (SASS) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        type_in_element_by_name(self.driver, 'reportDate', report_parameters['census_date'])
        
        select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        select_visible_text_in_element_by_name(self.driver, 'schoolGroup', 
            '[No Group Selected]') 

        # Submit report
        click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)

    # TODO: Refactor CRSC/CRSE to be a shared function rather than two totally separate ones
    def _download_fall2_report_for_course_section_records_crse(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Course Section Enrollment (CRSE) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_select_visible_text_in_element_by_id(self.driver, 'submission', 
            report_parameters['submission_type'])

        # Date fields need time to appear
        time.sleep(1) 
        helper_select_visible_text_in_element_by_id(self.driver, 'useTracks', 
            report_parameters['use_tracks'])
        helper_type_in_element_by_id(self.driver, 'censusDate', 
            report_parameters['census_date'])
        helper_select_visible_text_in_element_by_id(self.driver, 'bypassValidation', 
            'No' if validation_only_run else 'Yes')
        # The below indicates to not "Include Records For Course Code 1000"
        helper_select_visible_text_in_element_by_id(self.driver, 'selectCourseCode', 'No') 

        # Submit report
        helper_click_element_by_id(self.driver, 'submitReportSDKRuntimeParams')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_reportworks(destination_directory_path, 
            file_postfix)

    # TODO: Make SCSC and SCSE a shared function
    def _download_fall2_report_for_student_course_section_records_scse(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        ps_school_subdistrict_name:str, validation_only_run: bool=False):
        """
        Switches to the Student Course Enrollment (SCSE) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_select_visible_text_in_element_by_name(self.driver, 'submission', 
            report_parameters['submission_type'])
        helper_select_visible_text_in_element_by_name(self.driver, 'term', 
            report_parameters['term'])

        helper_type_in_element_by_name(self.driver, 'censusDate', 
            report_parameters['census_date'])

        helper_select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        helper_select_visible_text_in_element_by_name(self.driver, 'selectCourseCode', 
            'No') # Do not "Include Records For Course Code 1000"

        helper_select_visible_text_in_element_by_name(self.driver, 'subDistrict', 
            ps_school_subdistrict_name)

        # Submit report
        helper_click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)
    
    # EOY Reports ######################

    def _download_eoy_report_for_student_incident_records_sinc(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Student Incident Records (SINC) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_type_in_element_by_id(self.driver, 'reportStartDate', 
            report_parameters['report_start_date'])
        helper_type_in_element_by_id(self.driver, 'reportEndDate', 
            report_parameters['report_end_date'])
        helper_select_visible_text_in_element_by_id(self.driver, 'reportMode', 
            'Submission mode') # Only 'Submission mode' is supported by this tool
        time.sleep(1) # Give page time to react
        helper_select_visible_text_in_element_by_id(self.driver, 'bypassValidation', 
            'No' if validation_only_run else 'Yes')

        # Submit report
        helper_click_element_by_id(self.driver, 'submitReportSDKRuntimeParams')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_reportworks(destination_directory_path, 
            file_postfix)

    def _download_eoy_report_for_student_incident_results_records_sirs_or_student_offense_records_soff(
        self, file_postfix: str, destination_directory_path: str, ps_report_link_text: str, 
        report_parameters: dict, validation_only_run: bool=False):
        """
        Switches to the Student Incident Results Records (SIRS) or Student Offense Records (SOFF) 
        report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_type_in_element_by_id(self.driver, 'reportStartDate', 
            report_parameters['report_start_date'])
        helper_type_in_element_by_id(self.driver, 'reportEndDate', 
            report_parameters['report_end_date'])
        helper_select_visible_text_in_element_by_id(self.driver, 'bypassValidation', 
            'No' if validation_only_run else 'Yes')
        
        # Submit report
        helper_click_element_by_id(self.driver, 'submitReportSDKRuntimeParams')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_reportworks(destination_directory_path, 
            file_postfix)

    def _download_eoy_report_for_student_absence_summary_stas(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Student Absence Summary (STAS) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_type_in_element_by_name(self.driver, 'StartDate', 
            report_parameters['report_start_date'])
        helper_type_in_element_by_name(self.driver, 'EndDate', 
            report_parameters['report_end_date'])
        # Below 'adaFlag' defaults to 'Yes'
        helper_select_visible_text_in_element_by_name(self.driver, 'adaFlag', 'Yes') 
        helper_select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        # Below defaults to "No Group Selected" because school should already be chosen
        helper_select_visible_text_in_element_by_name(self.driver, 'schoolGroup', 
            '[No Group Selected]') 
        
        # Submit report
        helper_click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)

    def _download_student_program_records_sprg(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Student Program Records (SPRG) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_type_in_element_by_name(self.driver, 'startDate', 
            report_parameters['report_start_date'])
        helper_type_in_element_by_name(self.driver, 'endDate', 
            report_parameters['report_end_date'])
        helper_select_visible_text_in_element_by_name(self.driver, 'selectProgs', 
            report_parameters['submission_type'])

        # I believe 'Non-submission mode (all records)' is the one we want to use for repeated 
        #   submissions, but there's also 'Replacement Submission Mode' which has a different flow 
        #   after clicking Submit.
        helper_select_visible_text_in_element_by_name(self.driver, 'submissionMode', 
            'Non-submission mode (all records)') 

        helper_select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        # Below defaults to "No Group Selected" because school should already be chosen
        helper_select_visible_text_in_element_by_name(self.driver, 'schoolGroup', 
            '[No Group Selected]')

        # Submit report
        helper_click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)

    def _download_eoy_report_for_student_program_records_sprg(self, **kwargs):
        print("_download_eoy_report_for_student_program_records_sprg deprecated; use _download_student_program_records_sprg")
        return self._download_student_program_records_sprg(**kwargs)
    
    
    # TODO: Refactor CRSC/CRSE to be a shared function rather than two totally separate ones
    def _download_eoy_report_for_course_section_records_crsc(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        validation_only_run: bool=False):
        """
        Switches to the Course Section Completion (CRSC) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_select_visible_text_in_element_by_id(self.driver, 'submission', 
            report_parameters['submission_type'])

        # Date fields need time to appear
        time.sleep(1) 
        helper_type_in_element_by_id(self.driver, 'startDate', 
            report_parameters['report_start_date'])
        helper_type_in_element_by_id(self.driver, 'endDate', 
            report_parameters['report_end_date'])
        helper_select_visible_text_in_element_by_id(self.driver, 'bypassValidation', 
            'No' if validation_only_run else 'Yes')
        # The below indicates to not "Include Records For Course Code 1000"
        helper_select_visible_text_in_element_by_id(self.driver, 'selectCourseCode', 'No') 

        # Submit report
        helper_click_element_by_id(self.driver, 'submitReportSDKRuntimeParams')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_reportworks(destination_directory_path, 
            file_postfix)
    
    # TODO: Make SCSC and SCSE a shared function
    def _download_eoy_report_for_student_course_section_records_scsc(self, file_postfix: str, 
        destination_directory_path: str, ps_report_link_text: str, report_parameters: dict, 
        ps_school_subdistrict_name:str, validation_only_run: bool=False):
        """
        Switches to the Student Course Completion (SCSC) report in PowerSchool and downloads it.
        """
        self.navigate_to_specific_state_report(ps_report_link_text)
        
        # Enter specific parameters for this report
        helper_select_visible_text_in_element_by_name(self.driver, 'submission', 
            report_parameters['submission_type'])
        helper_type_in_element_by_name(self.driver, 'storeCodeList', 
            report_parameters['eoy_store_code_list'])
        # Below defaults to 'No' for 'Extract Credits for Grades 7 and 8' 
        # TODO: Research if this is correct
        helper_select_visible_text_in_element_by_name(self.driver, 'msExtract', 'No') 

        helper_type_in_element_by_name(self.driver, 'startDate', 
            report_parameters['report_start_date'])
        helper_type_in_element_by_name(self.driver, 'endDate', 
            report_parameters['report_end_date'])

        helper_select_visible_text_in_element_by_name(self.driver, 'bypass_validation', 
            'No' if validation_only_run else 'Yes')
        helper_select_visible_text_in_element_by_name(self.driver, 'selectCourseCode', 
            'No') # Do not "Include Records For Course Code 1000"

        helper_select_visible_text_in_element_by_name(self.driver, 'subDistrict', 
            ps_school_subdistrict_name)

        # Submit report
        helper_click_element_by_id(self.driver, 'btnSubmit')

        # Download report zipfile
        return self.download_latest_report_from_report_queue_system(destination_directory_path, 
            file_postfix)
    
    CALPADS_REPORT_TYPES = {
        'CRSC': {
            'ps_title': 'Course Section Records',
            'function': _download_eoy_report_for_course_section_records_crsc,
        },
        'CRSE': {
            'ps_title': 'Course Section Records',
            'function': _download_fall2_report_for_course_section_records_crse,
        },
        'SASS': {
            'ps_title': 'Staff Assignment Records',
            'function': _download_fall_2_report_for_staff_assignment_records_sass,
        },
        'SCSC': {
            'ps_title': 'Student Course Section Records',
            'function': _download_eoy_report_for_student_course_section_records_scsc,
        },
        'SCSE': {
            'ps_title': 'Student Course Section Records',
            'function': _download_fall2_report_for_student_course_section_records_scse,
        },
        'SDEM': {
            'ps_title': 'Staff Demographic Records',
            'function': _download_fall_2_report_for_staff_demographics_records_sdem,
        },    
        'SELA': {
            'ps_title': 'Student EL Acquisition Records',
            'function': _download_all_year_report_for_student_english_language_acquisition_records_sela,
        },
        'SENR': {
            'ps_title': 'SSID Enrollment Records',
            'function': _download_all_year_report_for_ssid_enrollment_records_senr,
        },
        'SINC': {
            'ps_title': 'Student Incident Records (SINC)',
            'function': _download_eoy_report_for_student_incident_records_sinc,
        },
        'SINF': {
            'ps_title': 'Student Information Records',
            'function': _download_all_year_report_for_student_information_records_sinf,
        },
        'SIRS': {
            'ps_title': 'Student Incident Results Records (SIRS)',
            'function': _download_eoy_report_for_student_incident_results_records_sirs_or_student_offense_records_soff,
        },
        'SOFF': {
            'ps_title': 'Student Offense Records (SOFF)',
            'function': _download_eoy_report_for_student_incident_results_records_sirs_or_student_offense_records_soff,
        },
        'SPRG': {
            'ps_title': 'Student Program Records',
            'function': _download_student_program_records_sprg,
        },
        'STAS': {
            'ps_title': 'Student Absence Summary',
            'function': _download_eoy_report_for_student_absence_summary_stas,
        },
    }
    
# Helper Functions #################

#    Note: These functions are intentionally not in the class, because one may want to access
#    them separately from downloading files from PowerSchool.

def swap_value_in_column_of_calpads_file(file_path: str, column_names_of_file: list, 
    column_name_for_swap: str, existing_value: str, new_value: str, encoding: str='utf-8'):
    """
    Opens a CALPADS upload file, looks for the existing_value in the provided column and replaces it
    with the new_value. Writes the new file to the same folder but with "_modified" added to the
    filename. Returns the new file path.

    Parameters:
    file_path: The path of the file that needs to be modified.
    column_names_of_file: A list storing the CALPADS column names for the relevant file. Needs to be 
        in order and exactly matching the number of columns in the file. File specifications available at: https://www.cde.ca.gov/ds/sp/cl/systemdocs.asp
    column_name_for_swap: The name of the column being modified.
    existing_value: The existing value in that column that needs to be swapped out.
    new_value: The value that will replace the existing_value.

    Returns:
    str: The path of the modified file
    """

    df_to_edit = pd.read_csv(file_path, sep='^', header=None, names=column_names_of_file, dtype=str, encoding=encoding)

    df_to_edit.loc[df_to_edit[column_name_for_swap] == existing_value, column_name_for_swap] = new_value

    df_to_edit.fillna('', inplace=True)

    updated_file_path = file_path.replace('.txt', '_modified.txt')
    df_to_edit.to_csv(updated_file_path, sep='^', header=False, index=False, na_rep='')

    return updated_file_path

def remove_sela_records_beginning_before_report_start_date(sela_file_path: str, 
    report_start_date: str):
    """
    Take the SELA file at the provided path, filter so it contains only the
    records with an ELA status start date greater than or equal to the report
    start date, and return the path to the modified file.
    """

    # Load existing upload file
    df_to_edit = pd.read_csv(sela_file_path, sep='^', header=None, names=SELA_COLUMN_NAMES, dtype=str, encoding='cp1252')

    # Convert report_start_date to a string in the format of 'YYYYMMDD' to match the upload file column
    report_start_date_object = datetime.strptime(report_start_date, '%m/%d/%Y')
    reformatted_report_start_date_string = report_start_date_object.strftime('%Y%m%d')

    # Keep only records greater than or equal to the report_start_date
    filtered_df = df_to_edit[df_to_edit['English Language Acquisition Status Start Date'] >= reformatted_report_start_date_string]

    # Output new upload file
    updated_file_path = sela_file_path.replace('.txt', '_modified.txt')
    filtered_df.to_csv(updated_file_path, sep='^', header=False, index=False, na_rep='')

    return updated_file_path